/*
** pork_command.c - interface to commands typed by the user
** Copyright (C) 2002-2003 Ryan McCabe <ryan@numb.org>
** Copyright (C) 2002-2003 Amber Adams <amber@ojnk.net>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <config.h>

#include <unistd.h>
#include <dirent.h>
#include <ncurses.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>

#include <libfaim/aim.h>
#include <libfaim/faimconfig.h>

#include <pork.h>
#include <pork_missing.h>
#include <pork_util.h>
#include <pork_list.h>
#include <pork_buddy.h>
#include <pork_set.h>
#include <pork_swindow.h>
#include <pork_imwindow.h>
#include <pork_slist.h>
#include <pork_buddy_list.h>
#include <pork_acct.h>
#include <pork_cstr.h>
#include <pork_misc.h>
#include <pork_html.h>
#include <pork_input.h>
#include <pork_bind.h>
#include <pork_events.h>
#include <pork_screen.h>
#include <pork_screen_cmd.h>
#include <pork_aim.h>
#include <pork_status.h>
#include <pork_alias.h>
#include <pork_opt.h>
#include <pork_conf.h>
#include <pork_perl.h>
#include <pork_timer.h>
#include <pork_command.h>
#include <pork_command_defs.h>
#include <pork_help.h>

extern struct screen screen;

static void print_binding(void *data, void *nothing);
static void print_alias(void *data, void *nothing);
static int cmd_compare(const void *l, const void *r);
static void print_timer(void *data, void *nothing);
static int run_one_command(char *str, u_int32_t set);

/*
** Note that the "struct command" arrays are arranged in alphabetical
** order.  They have to be like that.
*/

/*
** The / command set.
*/

static const struct command command[] = {
	{ "",			cmd_send			},
	{ "alias",		cmd_alias			},
	{ "away",		cmd_away			},
	{ "bind",		cmd_bind			},
	{ "blist",		cmd_blist			},
	{ "buddy",		cmd_buddy			},
	{ "chat",		cmd_chat			},
	{ "compose",	cmd_compose			},
	{ "connect",	cmd_connect			},
	{ "disconnect", cmd_disconnect		},
	{ "echo",		cmd_echo			},
	{ "email",		cmd_email			},
	{ "eval",		cmd_eval			},
	{ "event",		cmd_event			},
	{ "help",		cmd_help			},
	{ "history",	cmd_history			},
	{ "idle",		cmd_idle			},
	{ "input",		cmd_input			},
	{ "load",		cmd_load			},
	{ "msg",		cmd_msg				},
	{ "passwd",		cmd_passwd			},
	{ "perl",		cmd_perl			},
	{ "perl_dump",	cmd_perl_dump		},
	{ "perl_load",	cmd_perl_load		},
	{ "profile",	cmd_profile,		},
	{ "query",		cmd_query			},
	{ "quit",		cmd_quit			},
	{ "refresh",	cmd_refresh			},
	{ "save",		cmd_save			},
	{ "scroll",		cmd_scroll			},
	{ "set",		cmd_set				},
	{ "timer",		cmd_timer			},
	{ "unalias",	cmd_unalias			},
	{ "unbind",		cmd_unbind			},
	{ "win",		cmd_win				},
	{ "xecho",		cmd_xecho			},
};

/*
** /input commands
*/

static const struct command input_command[] = {
	{ "backspace",				cmd_input_bkspace			},
	{ "clear",					cmd_input_clear				},
	{ "clear_next_word",		cmd_input_clear_next		},
	{ "clear_prev_word",		cmd_input_clear_prev		},
	{ "clear_to_end",			cmd_input_clear_to_end		},
	{ "clear_to_start",			cmd_input_clear_to_start	},
	{ "delete",					cmd_input_delete			},
	{ "end",					cmd_input_end				},
	{ "focus_next",				cmd_input_focus_next		},
	{ "insert",					cmd_input_insert			},
	{ "left",					cmd_input_left				},
	{ "next_word",				cmd_input_next_word			},
	{ "prev_word",				cmd_input_prev_word			},
	{ "prompt",					cmd_input_prompt			},
	{ "right",					cmd_input_right				},
	{ "send",					cmd_input_send				},
	{ "start",					cmd_input_start				},
};

USER_COMMAND(cmd_input_bkspace) {
	input_bkspace(cur_window()->input);
}

USER_COMMAND(cmd_input_clear) {
	input_clear_line(cur_window()->input);
}

USER_COMMAND(cmd_input_clear_prev) {
	input_clear_prev_word(cur_window()->input);
}

USER_COMMAND(cmd_input_clear_next) {
	input_clear_next_word(cur_window()->input);
}

USER_COMMAND(cmd_input_clear_to_end) {
	input_clear_to_end(cur_window()->input);
}

USER_COMMAND(cmd_input_focus_next) {
	struct imwindow *imwindow = cur_window();

	imwindow_switch_focus(imwindow);
	imwindow_blist_draw(imwindow);
}

USER_COMMAND(cmd_input_clear_to_start) {
	input_clear_to_start(cur_window()->input);
}

USER_COMMAND(cmd_input_delete) {
	input_delete(cur_window()->input);
}

USER_COMMAND(cmd_input_end) {
	input_end(cur_window()->input);
}

USER_COMMAND(cmd_input_insert) {
	if (args != NULL)
		input_insert_str(cur_window()->input, args);
}

USER_COMMAND(cmd_input_left) {
	input_move_left(cur_window()->input);
}

USER_COMMAND(cmd_input_prev_word) {
	input_prev_word(cur_window()->input);
}

USER_COMMAND(cmd_input_prompt) {
	if (args == NULL) {
		input_set_prompt(cur_window()->input, NULL);
		screen_status_msg("Input prompt cleared");
	} else {
		if (input_set_prompt(cur_window()->input, args) == -1)
			screen_err_msg("The requested prompt is too long for the screen");
		else {
			screen_status_msg("Input prompt set to %s", args);
		}
	}
}

USER_COMMAND(cmd_input_next_word) {
	input_next_word(cur_window()->input);
}

USER_COMMAND(cmd_input_right) {
	input_move_right(cur_window()->input);
}

USER_COMMAND(cmd_input_send) {
	struct imwindow *imwindow = cur_window();
	struct input *input = imwindow->input;
	struct aim_priv *priv = imwindow->priv;
	static int recursion;

	/*
	** This is kind of a hack, but it's necessary if the client
	** isn't to crash when someone types "/input send"
	*/

	if (recursion == 1 && args == NULL)
		return;

	recursion = 1;

	if (args != NULL) {
		xstrncpy(input->input_buf, args, sizeof(input->input_buf));
		input->len = strlen(input->input_buf);
	}

	if (input->len > 0) {
		char *input_str = xstrdup(input->input_buf);

		input_history_add(input);
		input_clear_line(input);

		if (do_event(priv->events, EVENT_SEND_LINE, input_str, priv->username)
			&& opt_get_bool(OPT_I_KNOW_WHAT_IM_DOING))
		{
			recursion = 0;
			free(input_str);
			return;
		}

		if (*input_str == opt_get_char(OPT_CMDCHARS))
			run_command(&input_str[1]);
		else
			cmd_send(input_str);

		free(input_str);
	}

	recursion = 0;
}

USER_COMMAND(cmd_input_start) {
	input_home(cur_window()->input);
}

/*
** /scroll commands
*/

static const struct command scroll_command[] = {
	{ "by",				cmd_scroll_by			},
	{ "down",			cmd_scroll_down			},
	{ "end",			cmd_scroll_end			},
	{ "page_down",		cmd_scroll_pgdown		},
	{ "page_up",		cmd_scroll_pgup			},
	{ "start",			cmd_scroll_start		},
	{ "up",				cmd_scroll_up			},
};

USER_COMMAND(cmd_scroll_by) {
	int lines;

	if (args == NULL)
		return;

	if (str_to_int(args, &lines) != 0) {
		screen_err_msg("Bad number of lines: %s", args);
		return;
	}

	imwindow_scroll_by(cur_window(), lines);
}

USER_COMMAND(cmd_scroll_down) {
	imwindow_scroll_down(cur_window());
}

USER_COMMAND(cmd_scroll_end) {
	imwindow_scroll_end(cur_window());
}

USER_COMMAND(cmd_scroll_pgdown) {
	imwindow_scroll_page_down(cur_window());
}

USER_COMMAND(cmd_scroll_pgup) {
	imwindow_scroll_page_up(cur_window());
}

USER_COMMAND(cmd_scroll_start) {
	imwindow_scroll_start(cur_window());
}

USER_COMMAND(cmd_scroll_up) {
	imwindow_scroll_up(cur_window());
}

/*
** win commands.
*/

static const struct command window_command[] = {
	{ "bind",				cmd_win_bind		},
	{ "clear",				cmd_win_clear		},
	{ "close",				cmd_win_close		},
	{ "erase",				cmd_win_erase		},
	{ "list",				cmd_win_list		},
	{ "next",				cmd_win_next		},
	{ "prev",				cmd_win_prev		},
	{ "rename",				cmd_win_rename		},
	{ "renumber",			cmd_win_renumber	},
	{ "set",				cmd_win_set			},
	{ "swap",				cmd_win_swap		},
};

USER_COMMAND(cmd_win_bind) {
	struct imwindow *imwindow = cur_window();
	int ret;

	if (args == NULL || blank_str(args)) {
		if (imwindow->priv != NULL && imwindow->priv->username != NULL) {
			screen_output("This window is bound to account %s",
				imwindow->priv->username);
		} else
			screen_output("This window is bound to no account");

		return;
	}

	ret = imwindow_bind_acct(imwindow, args);
	if (ret == -1) {
		if (imwindow->type == TYPE_CHAT)
			screen_err_msg("You can't rebind chat windows");
		else
			screen_err_msg("Account %s isn't signed on", args);
	} else {
		screen_status_msg("This window is now bound to account %s",
			imwindow->priv->username);
	}
}

USER_COMMAND(cmd_win_clear) {
	imwindow_clear(cur_window());
}

USER_COMMAND(cmd_win_close) {
	screen_close_window(cur_window());
}

USER_COMMAND(cmd_win_erase) {
	imwindow_erase(cur_window());
}

USER_COMMAND(cmd_win_list) {
	dlist_t *cur;
	static const char *win_types[] = { "Status", "IM", "Chat Room" };

	screen_output("Window List:");

	cur = screen.window_list;
	do {
		struct imwindow *imwindow = cur->data;

		screen_output("[refnum %u] Name: %s, Type: %s, Target: %s",
			imwindow->refnum, imwindow->name,
			win_types[imwindow->type], imwindow->target);

		cur = cur->next;
	} while (cur != screen.window_list);
}

USER_COMMAND(cmd_win_next) {
	screen_cycle_fwd();
}

USER_COMMAND(cmd_win_prev) {
	screen_cycle_bak();
}

USER_COMMAND(cmd_win_rename) {
	imwindow_rename(cur_window(), xstrdup(args));
}

USER_COMMAND(cmd_win_renumber) {
	u_int32_t num;

	if (args == NULL) {
		screen_output("This is window %u", cur_window()->refnum);
		return;
	}

	if (str_to_uint(args, &num) != 0) {
		screen_err_msg("Bad window number: %s", args);
		return;
	}

	screen_renumber(cur_window(), num);
}

USER_COMMAND(cmd_win_set) {
	char *variable;
	char *value;
	int opt;

	variable = strsep(&args, " ");
	if (variable == NULL || blank_str(variable)) {
		wopt_print(cur_window());
		return;
	}

	opt = wopt_find(variable);
	allcaps(variable);
	if (opt == -1) {
		screen_err_msg("Unknown variable: %s", variable);
		return;
	}

	value = args;
	if (value == NULL || blank_str(value)) {
		wopt_print_var(cur_window(), opt, "is set to");
		return;
	}

	if (wopt_set(cur_window(), opt, value) == -1) {
		screen_err_msg("Bad argument for %s: %s", variable, value);
	} else {
		wopt_print_var(cur_window(), opt, "set to");
	}
}

/*
** History Manipulation.
*/

static const struct command history_command[] = {
	{ "clear",		cmd_history_clear	},
	{ "next",		cmd_history_next	},
	{ "prev",		cmd_history_prev	},
};

USER_COMMAND(cmd_history_clear) {
	input_history_clear(cur_window()->input);
}

USER_COMMAND(cmd_history_next) {
	input_history_next(cur_window()->input);
}

USER_COMMAND(cmd_history_prev) {
	input_history_prev(cur_window()->input);
}

/*
** /buddy commands
*/

static const struct command buddy_command[] = {
	{ "add",			cmd_buddy_add			},
	{ "add_group",		cmd_buddy_add_group		},
	{ "alias",			cmd_buddy_alias			},
	{ "awaymsg",		cmd_buddy_awaymsg		},
	{ "block",			cmd_buddy_block			},
	{ "clear_block",	cmd_buddy_clear_block	},
	{ "clear_permit",	cmd_buddy_clear_permit	},
	{ "list",			cmd_buddy_list			},
	{ "list_block",		cmd_buddy_list_block	},
	{ "list_permit",	cmd_buddy_list_permit	},
	{ "permit",			cmd_buddy_permit		},
	{ "privacy_mode",	cmd_buddy_privacy_mode	},
	{ "profile",		cmd_buddy_profile		},
	{ "remove",			cmd_buddy_remove		},
	{ "remove_group",	cmd_buddy_remove_group	},
	{ "remove_permit",	cmd_buddy_remove_permit	},
	{ "report_idle",	cmd_buddy_report_idle	},
	{ "search",			cmd_buddy_search		},
	{ "seen",			cmd_buddy_seen			},
	{ "unblock",		cmd_buddy_unblock		},
	{ "warn",			cmd_buddy_warn			},
	{ "warn_anon",		cmd_buddy_warn_anon		},
};

USER_COMMAND(cmd_buddy_add) {
	char *screen_name;
	char *group_name;
	struct bgroup *group;
	struct buddy *buddy;
	struct aim_priv *priv = cur_window()->priv;

	screen_name = strsep(&args, " ");
	if (screen_name == NULL || blank_str(screen_name)) {
		screen_err_msg("syntax is /buddy add <user> <group>");
		return;
	}

	buddy = buddy_find(priv, screen_name);
	if (buddy != NULL) {
		screen_err_msg("%s is already a member of the group %s",
			screen_name, buddy->group->name);
		return;
	}

	group_name = args;
	if (group_name == NULL || blank_str(group_name)) {
		screen_err_msg("syntax is /buddy add <user> <group>");
		return;
	}

	group = group_find(priv, group_name);
	if (group == NULL) {
		screen_err_msg("The group %s does not exist", group_name);
		return;
	}

	buddy_add(priv, screen_name, group, 1);
}

USER_COMMAND(cmd_buddy_block) {
	struct aim_priv *priv = cur_window()->priv;

	if (args == NULL || blank_str(args))
		return;

	if (buddy_add_block(priv, args, 1) == -1)
		screen_err_msg("%s is already on your block list", args);
	else
		screen_status_msg("%s added to block list", args);
}

USER_COMMAND(cmd_buddy_permit) {
	struct aim_priv *priv = cur_window()->priv;

	if (args == NULL || blank_str(args))
		return;

	if (buddy_add_permit(priv, args, 1) == -1)
		screen_err_msg("%s is already on your permit list", args);
	else
		screen_status_msg("%s added to permit list", args);
}

USER_COMMAND(cmd_buddy_add_group) {
	struct aim_priv *priv = cur_window()->priv;

	if (args == NULL || blank_str(args))
		return;

	group_add(priv, args);
}

USER_COMMAND(cmd_buddy_alias) {
	struct buddy *buddy;
	char *buddy_name;
	char *alias;
	struct aim_priv *priv = cur_window()->priv;
	struct imwindow *win;

	buddy_name = strsep(&args, " ");
	if (buddy_name == NULL)
		return;

	alias = args;
	if (alias == NULL || blank_str(alias))
		return;

	buddy = buddy_find(priv, buddy_name);
	if (buddy == NULL) {
		screen_err_msg("%s is not on your buddy list", buddy_name);
		return;
	}

	screen_status_msg("alias: %s -> %s", buddy_name, alias);
	buddy_alias(priv, buddy, alias, 1);

	/*
	** If there's a conversation window with this user open, we
	** should change its title.
	*/

	win = imwindow_find(buddy->nname);
	if (win != NULL)
		imwindow_rename(win, buddy->name);
}

USER_COMMAND(cmd_buddy_awaymsg) {
	char nname[NUSER_LEN];
	struct aim_priv *priv = cur_window()->priv;

	if (args == NULL || blank_str(args)) {
		aim_get_away_msg(priv, priv->username);
		return;
	}

	normalize(args, nname, sizeof(nname));
	aim_get_away_msg(priv, nname);
}

USER_COMMAND(cmd_buddy_clear_block) {
	buddy_clear_block(cur_window()->priv);
}

USER_COMMAND(cmd_buddy_clear_permit) {
	buddy_clear_permit(cur_window()->priv);
}

USER_COMMAND(cmd_buddy_privacy_mode) {
	struct aim_priv *priv = cur_window()->priv;
	int mode = -1;

	if (args != NULL)
		str_to_int(args, &mode);

	mode = aim_privacy_mode(priv, mode);
	screen_output("privacy mode for %s is %d", priv->username, mode);
}

USER_COMMAND(cmd_buddy_list) {
	struct imwindow *win = cur_window();
	struct aim_priv *priv = win->priv;
	struct buddy_pref *pref = priv->buddy_pref;
	dlist_t *gcur;

	if (pref == NULL)
		return;

	gcur = pref->group_list;
	if (gcur != NULL) {
		screen_win_msg(win, 0, 1, 1, "%s's buddy list: ",
			priv->username);
	}

	while (gcur != NULL) {
		struct bgroup *gr = gcur->data;
		dlist_t *bcur = gr->member_list;

		screen_win_msg(win, 0, 0, 1, "+ %s (%u/%u)",
			gr->name, gr->num_online, gr->num_members);

		while (bcur != NULL) {
			struct buddy *buddy = bcur->data;

			if (buddy->status != STATUS_OFFLINE) {
				char buddy_status[128];
				size_t i = 0;
				size_t len = sizeof(buddy_status);
				char *color_attr = "%G";

				buddy_status[0] = '\0';

				if (buddy->idle_time > 0) {
					char time_buf[64];
					int ret;

					color_attr = "%Y";

					time_to_str(buddy->idle_time, time_buf, sizeof(time_buf));
					ret = snprintf(buddy_status, len, "idle: %s ", time_buf);
					i += ret;
					len -= ret;
				}

				if (buddy->warn_level > 0) {
					snprintf(&buddy_status[i], len, "warn level: %d%%",
						buddy->warn_level);
				}

				screen_win_msg(win, 0, 0, 1,
					" %so%%x %s %s", color_attr, buddy->name, buddy_status);
			} else
				screen_win_msg(win, 0, 0, 1, " %%Ro%%x %s", buddy->name);

			bcur = bcur->next;
		}

		gcur = gcur->next;
	}
}

USER_COMMAND(cmd_buddy_list_permit) {
	struct imwindow *win = cur_window();
	struct aim_priv *priv = win->priv;
	struct buddy_pref *pref = priv->buddy_pref;
	dlist_t *cur;

	if (pref == NULL)
		return;

	cur = pref->permit_list;
	if (cur == NULL) {
		screen_win_msg(win, 0, 1, 1,
			"%s's permitted users list is empty", priv->username);
		return;
	}

	screen_win_msg(win, 0, 1, 1,
		"%s's permitted users list:", priv->username);

	while (cur != NULL) {
		screen_output(" - %s", (char *) cur->data);
		cur = cur->next;
	}
}

USER_COMMAND(cmd_buddy_list_block) {
	struct imwindow *win = cur_window();
	struct aim_priv *priv = win->priv;
	struct buddy_pref *pref = priv->buddy_pref;
	dlist_t *cur;

	if (pref == NULL)
		return;

	cur = pref->block_list;
	if (cur == NULL) {
		screen_win_msg(win, 0, 1, 1,
			"%s's blocked users list is empty", priv->username);
		return;
	}

	screen_win_msg(win, 0, 1, 1,
		"%s's blocked users list:", priv->username);

	while (cur != NULL) {
		screen_output(" - %s", (char *) cur->data);
		cur = cur->next;
	}
}

USER_COMMAND(cmd_buddy_profile) {
	struct imwindow *win = cur_window();
	struct aim_priv *priv = win->priv;

	if (args == NULL || blank_str(args)) {
		if (win->type == TYPE_IM)
			aim_whois(priv, win->target);
		else
			aim_whois(priv, priv->username);
	} else {
		char nname[NUSER_LEN];

		normalize(args, nname, sizeof(nname));
		aim_whois(priv, nname);
	}
}

USER_COMMAND(cmd_buddy_remove_permit) {
	struct aim_priv *priv = cur_window()->priv;

	if (args == NULL || blank_str(args))
		return;

	if (buddy_remove_permit(priv, args, 1) != 0)
		screen_err_msg("%s isn't on %s's permit list", args, priv->username);
	else {
		screen_status_msg("%s removed from %s's permit list",
			args, priv->username);
	}
}

USER_COMMAND(cmd_buddy_unblock) {
	struct aim_priv *priv = cur_window()->priv;

	if (args == NULL || blank_str(args))
		return;

	if (buddy_remove_block(priv, args, 1) != 0)
		screen_err_msg("%s isn't on %s's block list", args, priv->username);
	else {
		screen_status_msg("%s removed from %s's block list",
			args, priv->username);
	}
}

USER_COMMAND(cmd_buddy_remove) {
	struct aim_priv *priv = cur_window()->priv;

	if (args == NULL || blank_str(args))
		return;

	if (buddy_remove(priv, args, 1) != 0)
		screen_err_msg("%s isn't on %s's buddy list", args, priv->username);
	else {
		screen_status_msg("%s removed from %s's buddy list",
			args, priv->username);
	}
}

USER_COMMAND(cmd_buddy_remove_group) {
	struct aim_priv *priv = cur_window()->priv;

	if (args == NULL || blank_str(args))
		return;

	if (group_remove(priv, args, 1) != 0) {
		screen_err_msg("The group %s does not exist on %s's buddy list",
			args, priv->username);
	} else {
		screen_err_msg("The group %s was removed from %s's buddy list",
			args, priv->username);
	}
}

USER_COMMAND(cmd_buddy_report_idle) {
	struct aim_priv *priv = cur_window()->priv;

	if (args != NULL && priv->connected) {
		u_int32_t mode;

		if (str_to_uint(args, &mode) != 0) {
			screen_err_msg("Invalid number: %s", args);
			return;
		}

		aim_report_idle(priv, mode);
	}

	if (priv->buddy_pref == NULL)
		return;

	screen_status_msg("The reporting of idle time for %s is %s",
		priv->username,
		(priv->buddy_pref->report_idle ? "enabled" : "disabled"));
}

USER_COMMAND(cmd_buddy_search) {
	struct aim_priv *priv = cur_window()->priv;

	if (priv->connected && args != NULL)
		aim_usersearch_address(&priv->aim_session, priv->bos_conn, args);
}

USER_COMMAND(cmd_buddy_seen) {
	struct aim_priv *priv = cur_window()->priv;
	struct buddy *buddy;

	if (args == NULL || blank_str(args))
		return;

	buddy = buddy_find(priv, args);
	if (buddy == NULL) {
		screen_err_msg("%s is not on %s's buddy list", args, priv->username);
		return;
	}

	if (buddy->status != STATUS_OFFLINE) {
		screen_output("%s is currently online", args);
		return;
	}

	if (buddy->last_seen == 0) {
		screen_output("%s has never been seen online by %s",
			args, priv->username);
	} else {
		char *p;
		char timebuf[64];
		char timestr[64];
		u_int32_t time_diff = (u_int32_t) time(NULL) - buddy->last_seen;

		time_to_str(time_diff / 60 , timebuf, sizeof(timebuf));

		xstrncpy(timestr,
			asctime(localtime((time_t *) &buddy->last_seen)), sizeof(timestr));

		p = strchr(timestr, '\n');
		if (p != NULL)
			*p = '\0';

		screen_output("%s last saw %s online %s (%s ago)",
			priv->username, buddy->name, timestr, timebuf);
	}
}

USER_COMMAND(cmd_buddy_warn) {
	struct imwindow *win = cur_window();
	struct aim_priv *priv = win->priv;

	if (!priv->connected) {
		screen_err_msg("You're not connected");
		return;
	}

	if (args == NULL || blank_str(args)) {
		if (win->type == TYPE_IM)
			args = win->target;
		else
			return;
	}

	if (!do_event(priv->events, EVENT_SEND_WARN, args, 0, priv->username)) {
		aim_warn(priv, args);
		screen_status_msg("%s has warned %s", priv->username, args);
	}
}

USER_COMMAND(cmd_buddy_warn_anon) {
	struct imwindow *win = cur_window();
	struct aim_priv *priv = win->priv;

	if (!priv->connected) {
		screen_err_msg("You're not connected");
		return;
	}

	if (args == NULL || blank_str(args)) {
		if (win->type == TYPE_IM)
			args = win->target;
		else
			return;
	}

	if (!do_event(priv->events, EVENT_SEND_WARN, args, 1, priv->username)) {
		aim_warn_anon(priv, args);
		screen_status_msg("%s has warned %s anonymously", priv->username, args);
	}
}

/*
** /blist commands
*/

static const struct command blist_command[] = {
	{ "add_block",		cmd_blist_add_block		},
	{ "add_permit",		cmd_blist_add_permit	},
	{ "away",			cmd_blist_away			},
	{ "collapse",		cmd_blist_collapse		},
	{ "down",			cmd_blist_down			},
	{ "end",			cmd_blist_end			},
	{ "goto",			cmd_blist_goto			},
	{ "hide",			cmd_blist_hide			},
	{ "page_down",		cmd_blist_pgdown		},
	{ "page_up",		cmd_blist_pgup			},
	{ "profile",		cmd_blist_profile		},
	{ "refresh",		cmd_blist_refresh		},
	{ "remove",			cmd_blist_remove		},
	{ "remove_block",	cmd_blist_remove_block	},
	{ "remove_permit",	cmd_blist_remove_permit	},
	{ "select",			cmd_blist_select		},
	{ "show",			cmd_blist_show			},
	{ "start",			cmd_blist_start			},
	{ "up",				cmd_blist_up			},
	{ "warn",			cmd_blist_warn			},
	{ "warn_anon",		cmd_blist_warn_anon		},
	{ "width",			cmd_blist_width			},
};

USER_COMMAND(cmd_blist_add_block) {
	struct slist_cell *cell;
	struct buddy *buddy;
	struct blist *blist = cur_window()->priv->blist;

	if (blist == NULL) {
		screen_err_msg("Log in first");
		return;
	}

	cell = blist_get_cursor(blist);
	if (cell == NULL || cell->type == TYPE_LIST_CELL)
		return;

	buddy = cell->data;
	cmd_buddy_block(buddy->nname);
}

USER_COMMAND(cmd_blist_add_permit) {
	struct slist_cell *cell;
	struct buddy *buddy;
	struct blist *blist = cur_window()->priv->blist;

	if (blist == NULL) {
		screen_err_msg("Log in first");
		return;
	}

	cell = blist_get_cursor(blist);
	if (cell == NULL || cell->type == TYPE_LIST_CELL)
		return;

	buddy = cell->data;
	cmd_buddy_permit(buddy->nname);
}

USER_COMMAND(cmd_blist_away) {
	struct slist_cell *cell;
	struct buddy *buddy;
	struct blist *blist = cur_window()->priv->blist;

	if (blist == NULL) {
		screen_err_msg("Log in first");
		return;
	}

	cell = blist_get_cursor(blist);
	if (cell == NULL || cell->type == TYPE_LIST_CELL)
		return;

	buddy = cell->data;
	cmd_buddy_awaymsg(buddy->nname);
}

USER_COMMAND(cmd_blist_collapse) {
	struct slist_cell *cell;
	struct aim_priv *priv = cur_window()->priv;
	struct blist *blist = priv->blist;

	if (blist == NULL) {
		screen_err_msg("Log in first");
		return;
	}

	if (args == NULL)
		cell = blist_get_cursor(blist);
	else {
		dlist_t *node;
		struct bgroup *gr = group_find(priv, args);

		if (gr == NULL)
			return;

		node = gr->blist_line;
		if (node == NULL || node->data == NULL)
			return;
		cell = node->data;
	}

	if (cell == NULL || cell->type != TYPE_LIST_CELL)
		return;

	blist_collapse_group(blist, cell->data);
}

USER_COMMAND(cmd_blist_down) {
	struct blist *blist = cur_window()->priv->blist;

	if (blist == NULL) {
		screen_err_msg("Log in first");
		return;
	}

	blist_cursor_down(blist);
}

USER_COMMAND(cmd_blist_end) {
	struct blist *blist = cur_window()->priv->blist;

	if (blist == NULL) {
		screen_err_msg("Log in first");
		return;
	}

	blist_cursor_end(blist);
}

USER_COMMAND(cmd_blist_goto) {
	if (args != NULL)
		cmd_query(args);
	else {
		struct slist_cell *cell;
		struct buddy *buddy;
		struct blist *blist = cur_window()->priv->blist;

		if (blist == NULL) {
			screen_err_msg("Log in first");
			return;
		}

		cell = blist_get_cursor(blist);
		if (cell == NULL)
			return;

		if (cell->type != TYPE_LIST_CELL)
			return;

		buddy = cell->data;
		cmd_query(buddy->nname);
	}
}

USER_COMMAND(cmd_blist_hide) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;

	if (priv->blist == NULL) {
		screen_err_msg("Log in first");
		return;
	}

	imwindow_blist_hide(imwindow);
}

USER_COMMAND(cmd_blist_pgdown) {
	struct blist *blist = cur_window()->priv->blist;

	if (blist == NULL) {
		screen_err_msg("Log in first");
		return;
	}

	blist_cursor_pgdown(blist);
}

USER_COMMAND(cmd_blist_pgup) {
	struct blist *blist = cur_window()->priv->blist;

	if (blist == NULL) {
		screen_err_msg("Log in first");
		return;
	}

	blist_cursor_pgup(blist);
}

USER_COMMAND(cmd_blist_profile) {
	struct slist_cell *cell;
	struct buddy *buddy;
	struct blist *blist = cur_window()->priv->blist;

	if (blist == NULL) {
		screen_err_msg("Log in first");
		return;
	}

	cell = blist_get_cursor(blist);
	if (cell == NULL || cell->type == TYPE_LIST_CELL)
		return;

	buddy = cell->data;
	cmd_buddy_profile(buddy->nname);
}

USER_COMMAND(cmd_blist_refresh) {
	imwindow_blist_draw(cur_window());
}

USER_COMMAND(cmd_blist_remove) {
	struct slist_cell *cell;
	struct buddy *buddy;
	struct blist *blist = cur_window()->priv->blist;

	if (blist == NULL) {
		screen_err_msg("Log in first");
		return;
	}

	cell = blist_get_cursor(blist);
	if (cell == NULL || cell->type == TYPE_LIST_CELL)
		return;

	buddy = cell->data;
	cmd_buddy_remove(buddy->nname);
}

USER_COMMAND(cmd_blist_remove_block) {
	struct slist_cell *cell;
	struct buddy *buddy;
	struct blist *blist = cur_window()->priv->blist;

	if (blist == NULL) {
		screen_err_msg("Log in first");
		return;
	}

	cell = blist_get_cursor(blist);
	if (cell == NULL || cell->type == TYPE_LIST_CELL)
		return;

	buddy = cell->data;
	cmd_buddy_unblock(buddy->nname);
}

USER_COMMAND(cmd_blist_remove_permit) {
	struct slist_cell *cell;
	struct buddy *buddy;
	struct blist *blist = cur_window()->priv->blist;

	if (blist == NULL) {
		screen_err_msg("Log in first");
		return;
	}

	cell = blist_get_cursor(blist);
	if (cell == NULL || cell->type == TYPE_LIST_CELL)
		return;

	buddy = cell->data;
	cmd_buddy_remove_permit(buddy->nname);
}

USER_COMMAND(cmd_blist_select) {
	struct slist_cell *cell;
	struct blist *blist = cur_window()->priv->blist;

	if (blist == NULL) {
		screen_err_msg("Log in first");
		return;
	}

	cell = blist_get_cursor(blist);
	if (cell == NULL)
		return;

	if (cell->type == TYPE_LIST_CELL)
		cmd_blist_collapse(NULL);
	else {
		struct buddy *buddy = cell->data;

		cmd_blist_goto(buddy->nname);
	}
}

USER_COMMAND(cmd_blist_show) {
	struct imwindow *imwindow = cur_window();

	if (imwindow->priv->blist == NULL) {
		screen_err_msg("Log in first");
		return;
	}

	imwindow_blist_show(imwindow);
}

USER_COMMAND(cmd_blist_start) {
	struct blist *blist = cur_window()->priv->blist;

	if (blist == NULL) {
		screen_err_msg("Log in first");
		return;
	}

	blist_cursor_start(blist);
}

USER_COMMAND(cmd_blist_up) {
	struct blist *blist = cur_window()->priv->blist;

	if (blist == NULL) {
		screen_err_msg("Log in first");
		return;
	}

	blist_cursor_up(blist);
}

USER_COMMAND(cmd_blist_warn) {
	struct slist_cell *cell;
	struct buddy *buddy;
	struct blist *blist = cur_window()->priv->blist;

	if (blist == NULL) {
		screen_err_msg("Log in first");
		return;
	}

	cell = blist_get_cursor(blist);
	if (cell == NULL || cell->type == TYPE_LIST_CELL)
		return;

	buddy = cell->data;
	cmd_buddy_warn(buddy->nname);
}

USER_COMMAND(cmd_blist_warn_anon) {
	struct slist_cell *cell;
	struct buddy *buddy;
	struct blist *blist = cur_window()->priv->blist;

	if (blist == NULL) {
		screen_err_msg("Log in first");
		return;
	}

	cell = blist_get_cursor(blist);
	if (cell == NULL || cell->type == TYPE_LIST_CELL)
		return;

	buddy = cell->data;
	cmd_buddy_warn_anon(buddy->nname);
}

USER_COMMAND(cmd_blist_width) {
	u_int32_t new_len;
	struct blist *blist = cur_window()->priv->blist;

	if (blist == NULL) {
		screen_err_msg("Log in first");
		return;
	}

	if (args == NULL)
		return;

	if (str_to_uint(args, &new_len) != 0) {
		screen_err_msg("Error: invalid width: %s", args);
		return;
	}

	screen_blist_width(blist, new_len);
}

static const struct command timer_command[] = {
	{ "add",			cmd_timer_add			},
	{ "del",			cmd_timer_del			},
	{ "del_refnum",		cmd_timer_del_refnum	},
	{ "list",			cmd_timer_list			},
	{ "purge",			cmd_timer_purge			},
};

USER_COMMAND(cmd_timer_add) {
	char *p;
	u_int32_t interval;
	u_int32_t times;

	if (args == NULL)
		return;

	p = strsep(&args, " ");
	if (p == NULL)
		return;

	if (str_to_uint(p, &interval) != 0) {
		screen_err_msg("Invalid timer interval: %s", p);
		return;
	}

	p = strsep(&args, " ");
	if (p == NULL)
		return;

	if (str_to_uint(p, &times) != 0) {
		screen_err_msg("Invalid number of times to run: %s", p);
		return;
	}

	if (args == NULL || blank_str(args))
		return;

	timer_add(&screen.timer_list, args, interval, times);
}

USER_COMMAND(cmd_timer_del) {
	int ret;

	if (args == NULL)
		return;

	ret = timer_del(&screen.timer_list, args);
	if (ret == -1)
		screen_err_msg("No timer for \"%s\" was found", args);
	else
		screen_status_msg("Timer for \"%s\" was removed", args);
}

USER_COMMAND(cmd_timer_del_refnum) {
	u_int32_t refnum;
	int ret;

	if (args == NULL)
		return;

	if (str_to_uint(args, &refnum) != 0) {
		screen_err_msg("Bad timer refnum: %s", args);
		return;
	}

	ret = timer_del_refnum(&screen.timer_list, refnum);
	if (ret == -1)
		screen_err_msg("No timer with refnum %u was found", refnum);
	else
		screen_status_msg("Timer with refnum %u was removed", refnum);
}

USER_COMMAND(cmd_timer_list) {
	dlist_iterate(screen.timer_list, print_timer, NULL);
}

USER_COMMAND(cmd_timer_purge) {
	if (screen.timer_list != NULL) {
		timer_destroy(&screen.timer_list);
		screen_status_msg("All timers have been removed");
	}
}

/*
** /event commands
*/

static const struct command event_command[] = {
	{ "add",			cmd_event_add			},
	{ "del",			cmd_event_del			},
	{ "del_refnum",		cmd_event_del_refnum	},
	{ "list",			cmd_event_list			},
	{ "purge",			cmd_event_purge			},
};

USER_COMMAND(cmd_event_add) {
	char *event_type;
	u_int32_t refnum;
	struct event *events = cur_window()->priv->events;

	if (args == NULL || *args == '\0') {
		event_list(events, NULL);
		return;
	}

	event_type = strsep(&args, " ");
	if (event_type == NULL) {
		event_list(events, NULL);
		return;
	}

	allcaps(event_type);
	if (args == NULL) {
		event_list(events, event_type);
		return;
	}

	if (event_add(events, event_type, args, &refnum) != 0)
		screen_err_msg("Error adding handler for %s", event_type);
	else {
		screen_status_msg("Event handler %s installed for %s (refnum %u)",
			args, event_type, refnum);
	}
}

USER_COMMAND(cmd_event_del) {
	char *event_type;
	int ret;

	if (args == NULL)
		return;

	event_type = strsep(&args, " ");
	if (event_type == NULL)
		return;

	ret = event_del_type(cur_window()->priv->events, event_type, args);
	if (ret == 0) {
		if (args == NULL) {
			screen_status_msg("Successfully removed handler %s for %s",
				event_type, args);
		} else
			screen_status_msg("Successfully removed all handlers for %s", args);
	} else {
		if (args == NULL) {
			screen_err_msg("Error removing handler %s for %s",
				event_type, args);
		} else
			screen_err_msg("Error removing all handlers for %s", args);
	}
}

USER_COMMAND(cmd_event_del_refnum) {
	u_int32_t refnum;

	if (args == NULL)
		return;

	if (str_to_uint(args, &refnum) != 0) {
		screen_err_msg("Invalid event refnum: %s", args);
		return;
	}

	if (event_del_refnum(cur_window()->priv->events, refnum) != 0)
		screen_err_msg("Error deleting event refnum %s", args);
	else
		screen_status_msg("Event refnum %s was removed", args);
}

USER_COMMAND(cmd_event_list) {
	event_list(cur_window()->priv->events, args);
}

USER_COMMAND(cmd_event_purge) {
	event_purge(cur_window()->priv->events);
}

/*
** Chat commands
*/

static const struct command chat_command[] = {
	{ "ignore",				cmd_chat_ignore			},
	{ "invite",				cmd_chat_invite			},
	{ "join",				cmd_chat_join			},
	{ "leave",				cmd_chat_leave			},
	{ "list",				cmd_chat_list			},
	{ "send",				cmd_chat_send			},
	{ "unignore",			cmd_chat_unignore		},
	{ "who",				cmd_chat_who			},
};

USER_COMMAND(cmd_chat_ignore) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;
	struct chatroom *chat = NULL;
	char *first_arg;
	char *dest;
	char *p;

	if (args == NULL || !priv->connected)
		return;

	first_arg = strsep(&args, " ");

	/*
	** If the first arg is a chat room name, the second and third
	** arguments will be the target.
	**
	** If the first argument isn't a chat room name, assume the current
	** window's chat.  If the current window isn't a chat window, bail out.
	*/

	p = strchr(first_arg, '/');
	if (p == NULL) {
		chat = imwindow->chat;

		if (imwindow->type != TYPE_CHAT || chat == NULL) {
			screen_err_msg("You must specify a chat room if the current window is not a chat window");
			return;
		}

		dest = first_arg;
	} else {
		dlist_t *node;

		node = aim_find_chat_name(priv, first_arg);
		if (node == NULL) {
			screen_err_msg("%s is not joined to %s", priv->username, first_arg);
			return;
		}

		chat = node->data;
		dest = strsep(&args, " ");
	}

	if (dest == NULL) {
		screen_err_msg("You must specify a buddy to ignore");
		return;
	}

	if (aim_chat_ignore(dest, chat) != 0) {
		screen_err_msg("%s is not a member of %s/%d",
			dest, chat->title, chat->exchange);
	} else {
		screen_win_msg(chat->win, 1, 1, 1,
			"%s ignored %%M%s%%x in %%C%s/%u%%x",
			priv->username, dest, chat->title, chat->exchange);
	}
}

USER_COMMAND(cmd_chat_invite) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;
	struct chatroom *chat = NULL;
	char *first_arg;
	char *dest;
	char *p;

	if (args == NULL || !priv->connected)
		return;

	first_arg = strsep(&args, " ");

	p = strchr(first_arg, '/');
	if (p == NULL) {
		chat = imwindow->chat;

		if (imwindow->type != TYPE_CHAT || chat == NULL) {
			screen_err_msg("You must specify a chat room if the current window is not a chat window");
			return;
		}

		dest = first_arg;
	} else {
		dlist_t *node;

		node = aim_find_chat_name(priv, first_arg);
		if (node == NULL) {
			screen_err_msg("%s is not joined to %s", priv->username, first_arg);
			return;
		}

		chat = node->data;
		dest = strsep(&args, " ");
	}

	if (dest == NULL) {
		screen_err_msg("You must specify a user to invite");
		return;
	}

	if (!do_event(priv->events, EVENT_SEND_CHAT_INVITE, chat->title,
		chat->exchange, dest, args, priv->username))
	{
		if (args != NULL) {
			screen_win_msg(chat->win, 1, 1, 1,
				"%s invited %%M%s%%x to %%C%s/%u%%x (%s)",
				priv->username, dest, chat->title, chat->exchange, args);
		} else {
			screen_win_msg(chat->win, 1, 1, 1,
				"%s invited %%M%s%%x to %%C%s/%u%%x",
				priv->username, dest, chat->title, chat->exchange);
		}

		aim_chat_send_invite(priv, dest, args, chat);
	}
}

USER_COMMAND(cmd_chat_join) {
	struct aim_priv *priv = cur_window()->priv;
	char *exchange;
	int exchange_num = 4;
	struct imwindow *imwindow = NULL;
	dlist_t *node;

	if (args == NULL)
		return;

	if (!priv->connected) {
		screen_err_msg("You're not connected");
		return;
	}

	exchange = strchr(args, '/');
	if (exchange != NULL) {
		*exchange++ = '\0';

		if (str_to_int(exchange, &exchange_num) != 0) {
			screen_err_msg("Invalid exchange: %s", exchange);
			return;
		}
	}

	if (exchange_num < 4)
		exchange_num = 4;
	else if (exchange_num > 16)
		exchange_num = 16;

	node = aim_find_chat(priv, args, exchange_num);
	if (node != NULL)
		imwindow = ((struct chatroom *) node->data)->win;
	else {
		char buf[128];

		snprintf(buf, sizeof(buf), "%s/%d", args, exchange_num);

		imwindow = imwindow_find_chat_target(priv, buf);

		if (imwindow == NULL) {
			imwindow = screen_new_chat_window(priv, buf);
			if (imwindow == NULL) {
				screen_err_msg("Unable to create a new window for %s", buf);
				return;
			}
		}

		if (!do_event(priv->events, EVENT_SEND_CHAT_JOIN, args, exchange_num,
			priv->username))
		{
			aim_join_chatroom(priv, args, exchange_num);
		}
	}

	screen_goto_window(imwindow->refnum);
}

USER_COMMAND(cmd_chat_leave) {
	struct aim_priv *priv = cur_window()->priv;
	struct imwindow *win;
	struct chatroom *chat;

	if (!priv->connected) {
		screen_err_msg("You're not connected");
		return;
	}

	if (args == NULL || blank_str(args)) {
		win = cur_window();

		if (win->type != TYPE_CHAT)
			return;

		if (win->chat == NULL) {
			screen_close_window(win);
			return;
		}

		chat = win->chat;
	} else {
		dlist_t *cur;

		cur = aim_find_chat_name(priv, args);
		if (cur == NULL) {
			screen_err_msg("%s is not joined to %s", priv->username, args);
			return;
		}

		chat = cur->data;
		win = chat->win;
	}

	if (do_event(priv->events, EVENT_SEND_CHAT_LEAVE, chat->name,
		chat->title, chat->exchange, priv->username))
	{
		return;
	}

	if (aim_leave_chatroom(priv, chat->title, chat->exchange) != -1) {
		win->chat = NULL;
		screen_close_window(win);
	} else {
		screen_err_msg("Error leaving chat room %s/%d for %s",
			chat->title, chat->exchange, priv->username);
	}
}

USER_COMMAND(cmd_chat_list) {
	struct aim_priv *priv = cur_window()->priv;
	dlist_t *cur;

	cur = priv->chat_list;
	if (cur == NULL) {
		screen_output("%s is not joined to any chat rooms", priv->username);
		return;
	}

	screen_output("%s is joined to the following chat rooms", priv->username);
	do {
		struct chatroom *chat = cur->data;

		screen_output("  %s/%d in window refnum %u",
			chat->title, chat->exchange, chat->win->refnum);

		cur = cur->next;
	} while (cur != NULL);
}

USER_COMMAND(cmd_chat_send) {
	struct aim_priv *priv = cur_window()->priv;
	char *name;
	dlist_t *node;
	struct chatroom *chat;

	if (args == NULL)
		return;

	if (!priv->connected) {
		screen_err_msg("You're not connected");
		return;
	}

	name = strsep(&args, " ");
	if (name == NULL || args == NULL) {
		screen_err_msg("You must specify a chatroom and a message");
		return;
	}

	node = aim_find_chat_name(priv, name);
	if (node == NULL) {
		screen_err_msg("%s is not joined to %s", priv->username, name);
		return;
	}

	chat = node->data;

	if (!do_event(priv->events, EVENT_SEND_CHAT_MSG,
		chat->name, chat->title, chat->exchange, args,
		priv->username))
	{
		imwindow_send_im(chat->win, args);
		aim_chat_send(priv, args, chat);
	}
}

USER_COMMAND(cmd_chat_unignore) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;
	struct chatroom *chat = NULL;
	char *first_arg;
	char *dest;
	char *p;
	int ret;

	if (args == NULL || !priv->connected)
		return;

	first_arg = strsep(&args, " ");

	p = strchr(first_arg, '/');
	if (p == NULL) {
		chat = imwindow->chat;

		if (imwindow->type != TYPE_CHAT || chat == NULL) {
			screen_err_msg("You must specify a chat room if the current window is not a chat window");
			return;
		}

		dest = first_arg;
	} else {
		dlist_t *node;

		node = aim_find_chat_name(priv, first_arg);
		if (node == NULL) {
			screen_err_msg("%s is not joined to %s", priv->username, first_arg);
			return;
		}

		chat = node->data;
		dest = strsep(&args, " ");
	}

	if (dest == NULL) {
		screen_err_msg("You must specify a user to unignore");
		return;
	}

	ret = aim_chat_unignore(dest, chat);
	if (ret != 0) {
		screen_err_msg("%s is not ignoring %s in %s/%d",
			priv->username, dest, chat->title, chat->exchange);
	} else {
		screen_win_msg(chat->win, 1, 1, 1,
			"%s is no longer ignoring %%M%s%%x in %%C%s/%u%%x",
			priv->username, dest, chat->title, chat->exchange);
	}
}

USER_COMMAND(cmd_chat_who) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;
	struct chatroom *chat = NULL;

	if (!priv->connected)
		return;

	if (args == NULL) {
		chat = imwindow->chat;

		if (imwindow->type != TYPE_CHAT || chat == NULL) {
			screen_err_msg("You must specify a chat room if the current window is not a chat window");
			return;
		}
	} else {
		dlist_t *node;

		node = aim_find_chat_name(priv, args);
		if (node == NULL) {
			screen_err_msg("%s is not joined to %s", priv->username, args);
			return;
		}

		chat = node->data;
	}

	aim_chat_print_users(imwindow, chat);
}

static const struct {
	const struct command *set;
	size_t elem;
	char *type;
} command_set[] = {
	{	command,			array_elem(command),			"" 			},
	{	window_command,		array_elem(window_command),		"window "	},
	{	history_command,	array_elem(history_command),	"history "	},
	{	input_command,		array_elem(input_command),		"input "	},
	{	scroll_command,		array_elem(scroll_command),		"scroll "	},
	{	buddy_command,		array_elem(buddy_command),		"buddy "	},
	{	blist_command,		array_elem(blist_command),		"blist "	},
	{	timer_command,		array_elem(timer_command),		"timer "	},
	{	event_command,		array_elem(event_command),		"event "	},
	{	chat_command,		array_elem(chat_command),		"chat "		},
};

/*
** Main command set.
*/

USER_COMMAND(cmd_alias) {
	char *alias;
	char *str;

	alias = strsep(&args, " ");
	if (alias == NULL || blank_str(alias)) {
		hash_iterate(&screen.alias_hash, print_alias, NULL);
		return;
	}

	str = args;
	if (str == NULL || blank_str(str)) {
		struct alias *lalias = alias_find(&screen.alias_hash, alias);

		if (lalias != NULL) {
			screen_output("%s is aliased to %s%s",
				lalias->alias, lalias->orig,
				(lalias->args != NULL ? lalias->args : ""));
		} else
			screen_err_msg("There is no alias for %s", alias);

		return;
	}

	if (alias_add(&screen.alias_hash, alias, str) == 0) {
		struct alias *lalias = alias_find(&screen.alias_hash, alias);

		if (lalias != NULL) {
			screen_status_msg("%s is aliased to %s%s",
				lalias->alias, lalias->orig,
				(lalias->args != NULL ? lalias->args : ""));

			return;
		}
	}

	screen_err_msg("Error adding alias for %s", alias);
}

USER_COMMAND(cmd_away) {
	struct aim_priv *priv = cur_window()->priv;

	if (!priv->connected) {
		screen_err_msg("You're not connected");
		return;
	}

	if (!do_event(priv->events, EVENT_SEND_AWAY, args, priv->username))
		aim_set_away(priv, args);
}

USER_COMMAND(cmd_bind) {
	int key;
	char *func;
	char *key_str;
	struct key_binds *target_binds = cur_window()->active_binds;
	struct binding *binding;

	key_str = strsep(&args, " ");
	if (key_str == NULL || blank_str(key_str)) {
		hash_iterate(&target_binds->hash, print_binding, NULL);
		return;
	}

	if (key_str[0] == '-' && key_str[1] != '\0') {
		if (!strcasecmp(key_str, "-b") || !strcasecmp(key_str, "-buddy"))
			target_binds = &screen.binds.blist;
		else if (!strcasecmp(key_str, "-m") || !strcasecmp(key_str, "-main"))
			target_binds = &screen.binds.main;
		else {
			screen_err_msg("Bad bind flag: %s", key_str);
			return;
		}

		key_str = strsep(&args, " ");

		if (key_str == NULL || blank_str(key_str)) {
			hash_iterate(&target_binds->hash, print_binding, NULL);
			return;
		}
	}

	key = bind_get_keycode(key_str);
	if (key == -1) {
		screen_err_msg("Bad keycode: %s", key_str);
		return;
	}

	func = args;
	if (func != NULL) {
		if (*func == opt_get_char(OPT_CMDCHARS) && *(func + 1) != '\0')
			func++;
		if (blank_str(func))
			func = NULL;
	}

	if (func == NULL) {
		binding = bind_find(target_binds, key);
		if (binding != NULL)
			screen_output("%s is bound to %s", key_str, binding->binding);
		else
			screen_output("%s is not bound", key_str);

		return;
	}

	bind_add(target_binds, key, func);
	binding = bind_find(target_binds, key);
	if (binding != NULL) {
		screen_status_msg("%s is bound to %s", key_str, binding->binding);
		return;
	}

	screen_err_msg("Error binding %s", key_str);
}

USER_COMMAND(cmd_compose) {
	// XXX finish this
	return;
}

USER_COMMAND(cmd_connect) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv;
	dlist_t *node;

	if (args != NULL && !blank_str(args)) {
		struct pork_acct *acct;

		acct = pork_acct_add(args);
		if (acct == NULL) {
			screen_err_msg("%s is already connected", args);
			return;
		}

		priv = acct->priv;
		if (imwindow->priv == screen.null_priv) {
			cmd_win_bind(acct->username);

			if (!imwindow->blist_visible) {
				if (wopt_get_bool(imwindow->opts, WOPT_SHOW_BLIST))
					wopt_set(imwindow, WOPT_SHOW_BLIST, "1");
			}

		}
	} else {
		if (imwindow->priv->connected) {
			cmd_win_bind(imwindow->priv->username);
			return;
		}

		priv = imwindow->priv;
	}

	if (!priv->can_connect) {
		screen_err_msg("You must specify a screen name before connecting");
		return;
	}

	if (priv->passwd == NULL) {
		char buf[128];

		screen_get_passwd(buf, sizeof(buf));
		if (buf[0] == '\0') {
			screen_err_msg("There was an error reading your password");
			goto out_fail;
		}

		priv->passwd = xstrdup(buf);
		memset(buf, 0, sizeof(buf));
	}

	if (aim_login(priv) == -1) {
		screen_err_msg("Unable to login as %s", priv->username);
		goto out_fail;
	}

	return;

out_fail:
	node = pork_acct_find(priv->username);
	if (node != NULL)
		pork_acct_del(node);
}

USER_COMMAND(cmd_echo) {
	if (args != NULL)
		screen_win_msg(cur_window(), 0, 0, 1, "%s", args);
}

USER_COMMAND(cmd_eval) {
	if (args != NULL)
		run_mcommand(args);
}

USER_COMMAND(cmd_disconnect) {
	struct aim_priv *priv = cur_window()->priv;
	char *dest;
	dlist_t *node;
	struct pork_acct *acct;

	if (args == NULL || blank_str(args))
		dest = priv->username;
	else
		dest = args;

	node = pork_acct_find(dest);
	if (node == NULL) {
		screen_err_msg("Account %s is not logged in", dest);
		return;
	}

	acct = node->data;
	if (!acct->priv->can_connect) {
		screen_err_msg("You cannot sign %s off", acct->priv->username);
		return;
	}

	pork_acct_del(node);
}

USER_COMMAND(cmd_help) {
	char *section;

	if (args == NULL) {
		char buf[8192];

		if (pork_help_get_cmds("main", buf, sizeof(buf)) != -1) {
			screen_output("Help for the following commands is available:", sizeof(buf));
			screen_win_msg(cur_window(), 0, 0, 1, "   %s", buf);
		} else
			screen_err_msg("Error: Can't find the help files");

		return;
	}

	section = strsep(&args, " ");
	if (section == NULL) {
		screen_err_msg("Error: Can't find the help files");
		return;
	}

	if (args == NULL) {
		char buf[8192];

		if (pork_help_print("main", section) == -1) {
			screen_err_msg("Help: Error: No such command or section: %s",
				section);
		} else {
			if (pork_help_get_cmds(section, buf, sizeof(buf)) != -1) {
				screen_win_msg(cur_window(), 0, 0, 0, " ");
				allcaps(section);
				screen_win_msg(cur_window(), 0, 0, 1, "%%W%s COMMANDS",
					section);
				screen_win_msg(cur_window(), 0, 0, 1, "    %s", buf);
				screen_output("Type /help %s <command> for the help text for a particular %s command.",
					section, section);
			}
		}
	} else {
		if (pork_help_print(section, args) == -1) {
			screen_err_msg("Help: Error: No such command in section %s",
				section, args);
		}
	}
}

USER_COMMAND(cmd_idle) {
	u_int32_t idle_secs;
	struct aim_priv *priv = cur_window()->priv;

	if (!priv->connected) {
		screen_err_msg("You're not connected");
		return;
	}

	if (args == NULL) {
		if (!do_event(priv->events, EVENT_SEND_IDLE, 0, priv->username)) {
			aim_set_idle(priv, 0);
			screen_output("%s's idle time set to 0", priv->username);
		}

		return;
	}

	if (str_to_uint(args, &idle_secs) != 0) {
		screen_err_msg("Bad argument: %s", args);
		return;
	}

	if (!do_event(priv->events, EVENT_SEND_IDLE, idle_secs, priv->username)) {
		char timebuf[32];

		aim_set_idle(priv, idle_secs);
		time_to_str(idle_secs, timebuf, sizeof(timebuf));

		screen_output("%s's idle time set to %s", priv->username, timebuf);
	}
}

USER_COMMAND(cmd_load) {
	int quiet;

	if (args == NULL)
		return;

	quiet = screen_set_quiet(1);
	if (read_conf(args) != 0)
		screen_err_msg("Error reading %s: %s", args, strerror(errno));
	screen_set_quiet(quiet);
}

USER_COMMAND(cmd_msg) {
	struct imwindow *conv_window;
	char *target;
	char *msg;
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv;

	if (args == NULL)
		return;

	if (!imwindow->priv->connected) {
		screen_err_msg("You're not logged in");
		return;
	}

	target = strsep(&args, " ");
	if (target == NULL)
		return;

	msg = args;
	if (msg == NULL)
		return;

	conv_window = imwindow_find(target);
	if (conv_window == NULL || conv_window->type != TYPE_IM) {
		struct buddy *buddy = buddy_find(imwindow->priv, target);
		char *name;

		if (buddy != NULL)
			name = buddy->name;
		else
			name = target;

		conv_window = screen_new_window(imwindow->priv, target, name);
		if (conv_window == NULL)
			return;

		screen_goto_window(conv_window->refnum);
	}

	priv = conv_window->priv;
	if (!do_event(priv->events, EVENT_SEND_IM, target, msg, priv->username)) {
		aim_send_msg(priv, target, msg);
		imwindow_send_im(conv_window, msg);
	}
}

USER_COMMAND(cmd_query) {
	struct imwindow *imwindow = cur_window();

	if (args != NULL && !blank_str(args)) {
		struct imwindow *conv_window;

		conv_window = imwindow_find(args);
		if (conv_window == NULL) {
			struct buddy *buddy = buddy_find(imwindow->priv, args);
			char *name;

			if (buddy != NULL)
				name = buddy->name;
			else
				name = args;

			conv_window = screen_new_window(imwindow->priv, args, name);
		}

		screen_goto_window(conv_window->refnum);
	} else
		screen_close_window(imwindow);
}

USER_COMMAND(cmd_quit) {
	if (!do_event(cur_window()->priv->events, EVENT_QUIT))
		pork_exit(0, NULL);
}

USER_COMMAND(cmd_refresh) {
	screen_refresh();
}

USER_COMMAND(cmd_save) {
	if (save_global_config() == 0)
		screen_status_msg("Your configuration has been saved");
	else
		screen_err_msg("There was an error saving your configuration");
}

USER_COMMAND(cmd_send) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;

	if (args == NULL)
		return;

	if (!priv->connected) {
		if (imwindow->type != TYPE_STATUS)
			screen_err_msg("You're not connected");

		return;
	}

	if (imwindow->type == TYPE_IM) {
		if (!do_event(priv->events, EVENT_SEND_IM, imwindow->target, args, priv->username)) {
			imwindow_send_im(imwindow, args);
			aim_send_msg(priv, imwindow->target, args);
		}
	} else if (imwindow->type == TYPE_CHAT) {
		struct chatroom *chat = imwindow->chat;

		if (chat == NULL) {
			screen_err_msg("%s is not joined to this chat room",
				priv->username);
		} else {
			if (!do_event(priv->events, EVENT_SEND_CHAT_MSG,
				chat->name, chat->title, chat->exchange, args,
				priv->username))
			{
				imwindow_send_im(imwindow, args);
				aim_chat_send(priv, args, imwindow->chat);
			}
		}
	}
}

USER_COMMAND(cmd_unbind) {
	char *binding;
	struct imwindow *imwindow = cur_window();
	struct key_binds *target_binds = imwindow->active_binds;
	int c;

	binding = strsep(&args, " ");
	if (binding == NULL || blank_str(binding))
		return;

	if (binding[0] == '-' && binding[1] != '\0') {
		if (!strcasecmp(binding, "-b") || !strcasecmp(binding, "-buddy"))
			target_binds = &screen.binds.blist;
		else if (!strcasecmp(binding, "-m") || !strcasecmp(binding, "-main"))
			target_binds = &screen.binds.main;
		else {
			screen_err_msg("Bad unbind flag: %s", binding);
			return;
		}

		binding = strsep(&args, " ");

		if (binding == NULL || blank_str(binding))
			return;
	}

	c = bind_get_keycode(binding);
	if (c == -1) {
		screen_err_msg("Bad keycode: %s", binding);
		return;
	}

	if (bind_remove(target_binds, c) == -1)
		screen_err_msg("There is no binding for %s", binding);
	else
		screen_status_msg("Binding for %s removed", binding);
}

USER_COMMAND(cmd_unalias) {
	if (args == NULL)
		return;

	if (alias_remove(&screen.alias_hash, args) == -1)
		screen_err_msg("No such alias: %s", args);
	else
		screen_status_msg("Alias %s removed", args);
}

USER_COMMAND(cmd_email) {
	struct imwindow *imwindow = cur_window();

	if (!imwindow->priv->connected) {
		screen_err_msg("You're not connected");
		return;
	}
}

USER_COMMAND(cmd_passwd) {
	struct imwindow *imwindow = cur_window();

	if (!imwindow->priv->connected) {
		screen_err_msg("You're not connected");
		return;
	}
}

USER_COMMAND(cmd_perl) {
	size_t num_args = 0;
	char *p = args;
	char **perl_args;
	char *function;
	char *orig;
	size_t i = 0;

	if (args == NULL || blank_str(args))
		return;

	p = args;
	while (*p == ' ')
		p++;

	if (*p == '$')
		p++;
	args = p;
	orig = p;

	function = args;
	p = strchr(function, '(');
	if (p != NULL) {
		*p++ = '\0';
		args = p;
	}

	p = strchr(function, ' ');
	if (p != NULL) {
		*p = '\0';
		args = p + 1;
	}

	if (args == orig) {
		execute_perl(function, NULL);
		return;
	}

	p = args;
	while ((p = strchr(p, ',')) != NULL) {
		++num_args;
		++p;
	}
	num_args += 2;

	p = strchr(args, ')');
	if (p != NULL)
		*p = '\0';

	perl_args = xcalloc(num_args, sizeof(char *));

	while ((p = strsep(&args, ",")) != NULL) {
		char *s;

		while (*p == ' ' || *p == '\t')
			p++;

		s = strrchr(p, ' ');
		if (s != NULL && blank_str(s)) {
			while (*s == ' ')
				s--;
			*(s + 1) = '\0';
		}

		perl_args[i++] = p;
	}

	perl_args[i] = NULL;

	execute_perl(function, perl_args);
	free(perl_args);
}

/*
** This destroys the perl state.
** It has the effect of unloads all scripts.  All scripts should
** be catching the EVENT_UNLOAD event and doing any necessary cleanup
** when it happens.
*/

USER_COMMAND(cmd_perl_dump) {
	struct aim_priv *priv = cur_window()->priv;

	/*
	** If events are ever made per-user this will have to change.
	*/
	do_event(priv->events, EVENT_UNLOAD);

	perl_destroy();
	perl_init();
}

USER_COMMAND(cmd_perl_load) {
	int ret;

	if (args == NULL)
		return;

	ret = perl_load_file(args);
	if (ret != 0)
		screen_err_msg("Error: The file %s couldn't be loaded", args);
}

USER_COMMAND(cmd_profile) {
	struct aim_priv *priv = cur_window()->priv;

	if (!priv->connected) {
		screen_err_msg("You're not connected");
		return;
	}

	free(priv->profile);
	if (args == NULL)
		priv->profile = NULL;
	else
		priv->profile = xstrdup(args);

	if (!do_event(priv->events, EVENT_SEND_PROFILE, args, priv->username)) {
		aim_set_profile(priv, args);
		screen_output("Profile for %s was %s", priv->username,
			(args == NULL ? "cleared" : "set"));
	}
}

USER_COMMAND(cmd_win_swap) {
	u_int32_t num;

	if (args == NULL)
		return;

	if (str_to_uint(args, &num) != 0) {
		screen_err_msg("Invalid window refnum: %s", args);
		return;
	}

	if (screen_goto_window(num) != 0)
		screen_err_msg("No such window: %s", args);
}

USER_COMMAND(cmd_xecho) {
	bool show_banner = 0;
	bool show_ts = 0;
	bool use_color = 1;
	bool more_args = 1;
	struct imwindow *imwindow = cur_window();
	char *window_arg;
	char *commands;

	if (args == NULL)
		return;

	while (args != NULL && args[0] == '-' && args[1] != '\0' && more_args) {
		commands = strsep(&args, " ");

		switch (commands[1]) {
			case '-':
				more_args = 0;
				break;

			case 'w':
			case 'W':
				window_arg = strsep(&args, " ");
				if (window_arg != NULL && window_arg[0] != '\0') {
					u_int32_t refnum;

					if (str_to_uint(window_arg, &refnum) != 0) {
						screen_err_msg("Bad window argument: %s", window_arg);
						return;
					}

					imwindow = imwindow_find_refnum(refnum);
					if (imwindow == NULL) {
						screen_err_msg("No such window: %s", window_arg);
						return;
					}
				} else {
					screen_err_msg("No window was specified");
				}

				break;

			case 'n':
			case 'N':
				use_color = 0;
				break;

			case 't':
			case 'T':
				show_ts = 1;
				break;

			case 'b':
			case 'B':
				show_banner = 1;
				break;

			case 'X':
				args = text2html(args);

			default:
				more_args = 0;
				break;
		}
	}

	if (args != NULL && args[0] != '\0')
		screen_win_msg(imwindow, show_ts, show_banner, use_color, "%s", args);
}

USER_COMMAND(cmd_event) {
	if (args != NULL)
		run_one_command(args, CMDSET_EVENT);
	else
		run_one_command("list", CMDSET_EVENT);
}

USER_COMMAND(cmd_chat) {
	if (args != NULL)
		run_one_command(args, CMDSET_CHAT);
	else
		run_one_command("list", CMDSET_CHAT);
}

USER_COMMAND(cmd_win) {
	if (args != NULL)
		run_one_command(args, CMDSET_WIN);
	else
		run_one_command("list", CMDSET_WIN);
}

USER_COMMAND(cmd_buddy) {
	if (args != NULL)
		run_one_command(args, CMDSET_BUDDY);
	else
		run_one_command("list", CMDSET_BUDDY);
}

USER_COMMAND(cmd_blist) {
	if (args != NULL)
		run_one_command(args, CMDSET_BLIST);
}

USER_COMMAND(cmd_input) {
	if (args != NULL)
		run_one_command(args, CMDSET_INPUT);
}

USER_COMMAND(cmd_history) {
	if (args != NULL)
		run_one_command(args, CMDSET_HISTORY);
}

USER_COMMAND(cmd_scroll) {
	if (args != NULL)
		run_one_command(args, CMDSET_SCROLL);
}

USER_COMMAND(cmd_timer) {
	if (args != NULL)
		run_one_command(args, CMDSET_TIMER);
	else
		run_one_command("list", CMDSET_TIMER);
}

USER_COMMAND(cmd_set) {
	char *variable;
	char *value;
	int opt;

	variable = strsep(&args, " ");
	if (variable == NULL || blank_str(variable)) {
		opt_print();
		return;
	}

	allcaps(variable);
	opt = opt_find(variable);
	if (opt == -1) {
		screen_err_msg("Unknown variable: %s", variable);
		return;
	}

	value = args;
	if (value == NULL || blank_str(value)) {
		opt_print_var(opt, "is set to");
		return;
	}

	if (opt_set(opt, value) == -1) {
		screen_err_msg("Bad argument for %s: %s", variable, value);
	} else {
		opt_print_var(opt, "set to");
	}
}

inline int run_command(char *str) {
	return (run_one_command(str, CMDSET_MAIN));
}

int run_mcommand(char *str) {
	int i = 0;
	char *copystr = xstrdup(str);
	char *cmdstr = copystr;
	char *curcmd;

	curcmd = strsep(&cmdstr, ";");
	if (curcmd == NULL)
		i = run_one_command(cmdstr, CMDSET_MAIN);
	else {
		while (curcmd != NULL && i != -1) {
			char cmdchars = opt_get_char(OPT_CMDCHARS);

			while (*curcmd == ' ')
				curcmd++;

			while (*curcmd == cmdchars)
				curcmd++;

			i = run_one_command(curcmd, CMDSET_MAIN);
			curcmd = strsep(&cmdstr, ";");
		}
	}

	free(copystr);
	return (i);
}

static int run_one_command(char *str, u_int32_t set) {
	char *cmd_str;
	struct command *cmd;

	if (set == CMDSET_MAIN) {
		int ret;
		char *alias_str;

		ret = alias_resolve(&screen.alias_hash, str, &alias_str);
		if (ret == 0)
			str = alias_str;
		else if (ret == 1) {
			screen_err_msg("The alias chain is too long");
			return (-1);
		}
	}

	cmd_str = strsep(&str, " \t");

	cmd = bsearch(cmd_str, command_set[set].set,
			command_set[set].elem, sizeof(struct command), cmd_compare);

	if (cmd == NULL) {
		screen_err_msg("Unknown %scommand: %s",
			command_set[set].type, cmd_str);

		return (-1);
	}

	cmd->cmd(str);
	return (0);
}

static void print_binding(void *data, void *nothing __notused) {
	struct binding *binding = data;
	char key_name[32];

	bind_get_keyname(binding->key, key_name, sizeof(key_name));
	screen_output("%s is bound to %s", key_name, binding->binding);
}

static void print_alias(void *data, void *nothing __notused) {
	struct alias *alias = data;

	screen_output("%s is aliased to %s%s",
		alias->alias, alias->orig, (alias->args != NULL ? alias->args : ""));
}

static int cmd_compare(const void *l, const void *r) {
	char *key = (char *) l;
	struct command *cmd = (struct command *) r;

	return (strcasecmp(key, cmd->name));
}

static void print_timer(void *data, void *nothing __notused) {
	struct timer_entry *timer = data;

	screen_output("[refnum: %u] %d %u %s", timer->refnum,
		timer->interval, timer->times, timer->command);
}
